WITH Ada.Text_IO;
USE Ada.Text_IO;
WITH Ada.Float_Text_IO;
USE Ada.Float_Text_IO;
WITH Ada.Numerics.Elementary_Functions;
USE Ada.Numerics.Elementary_Functions;

PACKAGE BODY Fctmulti IS

   --------------Cette fonction permet de crer une matrice-----------------------
   -------------------------------------------------------------------------------
   FUNCTION Creer_Matrice (
         L : IN     Integer;
         C : IN     Integer)
     RETURN MATRICE IS
      M : Matrice;
   BEGIN
      M.L := L;
      M.C := C;
      RETURN M;
   END Creer_Matrice;
   -------------------------------------------------------------------------------



   --------------Cette fonction permet de crer un vecteur------------------------
   -------------------------------------------------------------------------------
   FUNCTION Creer_Vecteur (
         L : IN     Integer)
     RETURN VECTEUR IS
      V : Vecteur;
   BEGIN
      V.L := L;
      RETURN V;
   END Creer_Vecteur;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcule la moyenne de chaque colonne---------------
   ---------------------d'une matrice et retourne le ------------------------------
   ------------------------resultat dans un vecteur--------------------------------
   --------------------------------------------------------------------------------
   FUNCTION Moyenne_Colonne (
         M : IN     Matrice)
     RETURN Vecteur IS
      V : Vecteur;
   BEGIN
      V.L := M.C;
      FOR I IN 1 .. M.C LOOP
         V.Vect(I) := 0.0;
         FOR J IN 1 .. M.L LOOP
            V.Vect(I) := V.Vect(I) + M.Mat(J,I);
         END LOOP;
         V.Vect(I) := V.Vect(I)/ Float(M.L);
      END LOOP;
      RETURN V;
   END Moyenne_Colonne;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcule la moyenne des carres de chaque-------------
   ------------------colonne d'une matrice et retourne le --------------------------
   ------------------------resultat dans un vecteur---------------------------------
   ---------------------------------------------------------------------------------
   FUNCTION Moyenne_Carre_Colonne (
         M : IN     Matrice)
     RETURN Vecteur IS
      V : Vecteur;
   BEGIN
      V.L := M.C;
      FOR I IN 1 .. M.C LOOP
         V.Vect(I) := 0.0;
         FOR J IN 1 .. M.L LOOP
            V.Vect(I) := V.Vect(I) + M.Mat(J,I)**2;
         END LOOP;
         V.Vect(I) := V.Vect(I)/ Float(M.L);
      END LOOP;
      RETURN V;
   END Moyenne_Carre_Colonne;
   -------------------------------------------------------------------------------



   ------------------Cette fonction calcule la transpose---------------------------
   ------------------d'une matrice et retourne le rsultat -------------------------
   ---------------------------------------------------------------------------------
   FUNCTION Transposee (
         M : IN     Matrice)
     RETURN Matrice IS
      M1 : Matrice;
   BEGIN
      M1.L := M.C;
      M1.C := M.L;
      FOR I IN 1 .. M.L LOOP
         FOR J IN 1 .. M.C LOOP
            M1.Mat(J,I) := M.Mat(I,J);
         END LOOP;
      END LOOP;
      RETURN M1;
   END Transposee;
   -------------------------------------------------------------------------------



   ------------Cette fonction prend une matrice qconque en -------------
   ------------paramtre; on centre et rduit celle_ci p.r. -----------
   ------------chaque colonne pour obtenir M1. M2 est la transpose-----
   ------------de M1. Et on retourne le produit matriciel de M1 et------
   ------------M2 afin d'avoir une matrice dfinie positive.------------
   ---------------------------------------------------------------------
   FUNCTION Matrice_Centree_Reduite (
         M : IN     Matrice)
     RETURN Matrice IS
      M1,
      M2,
      M3 : Matrice;
      V1,
      V2 : Vecteur;
   BEGIN
      V1 := Moyenne_Colonne(M); -- On calcule la moyenne de chaque colonne

      V2 := Moyenne_Carre_Colonne(M); -- On calcule la moyenne des carres de chaque colonne

      M1 := M;
      FOR I IN 1 .. M.L LOOP
         FOR J IN 1 .. M.C LOOP

            M1.Mat(I,J) := (M.Mat(I,J) - V1.Vect(J))/Sqrt((V2.Vect(J)-V1.Vect(J)**2)); -- M1 est la centre rduite de M
         END LOOP;
      END LOOP;

      M2 := Transposee(M1); -- M2 est la transpose de M1

      M3 := Produit_Matrice(M2,M1); -- M3 est le produit de M1 et M2 : elle est dfinie positive

      RETURN M3;
   END Matrice_Centree_Reduite;
   -------------------------------------------------------------------------------





   --------------Cette fonction calcul le produit de deux--------------
   --------------matrices et retourne une nouvelle matrice---------------
   -------------------------------------------------------------------------------
   FUNCTION Produit_Matrice (
         M1 : IN     Matrice;
         M2 : IN     MATRICE)
     RETURN Matrice IS
      M : Matrice;
   BEGIN
      M.L := M1.L;
      M.C := M2.C;
      IF(M1.C/=M2.L)THEN
         RAISE Erreur_Dimension_Produit_Matrice;
      ELSE
         FOR I IN 1 .. M1.L LOOP
            FOR J IN 1 .. M2.C LOOP
               FOR K IN 1 .. M1.C LOOP
                  M.Mat(I,J) := M.Mat(I,J) + M1.Mat(I,K)*M2.Mat(K,J);
               END LOOP;
            END LOOP;
         END LOOP;
      END IF;
      RETURN M;
   END Produit_Matrice;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcul la somme de deux--------------
   --------------matrices et retourne une nouvelle matrice---------------
   -------------------------------------------------------------------------------
   FUNCTION Somme_Matrice (
         M1 : IN     Matrice;
         M2 : IN     MATRICE)
     RETURN Matrice IS
      M : Matrice;
   BEGIN
      M.L := M1.L;
      M.C := M2.C;
      IF(M1.L/=M2.L OR M1.C/=M2.C)THEN
         RAISE Erreur_Dimension_Produit_Matrice;
      ELSE
         FOR I IN 1 .. M.L LOOP
            FOR J IN 1 .. M.C LOOP
               M.Mat(I,J) := M1.Mat(I,J) + M2.Mat(I,J);
            END LOOP;
         END LOOP;
      END IF;
      RETURN M;
   END Somme_Matrice;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul le produit d'un scalaire--------------
   --------------et d'une matrice => on multiplie chaque case ---------------
   ----------------------------par le scalaire-------------------------------
   --------------------------------------------------------------------------
   FUNCTION Produit_Matrice_Scalaire (
         M : Matrice;
         S : Float)
     RETURN Matrice IS

      Mr : Matrice;
   BEGIN
      Mr.L := M.L;
      Mr.C := M.C;
      FOR I IN 1 .. M.L LOOP
         FOR J IN 1 .. M.C LOOP
            Mr.Mat(I,J) := M.Mat(I,J)*S;
         END LOOP;
      END LOOP;
      RETURN Mr;
   END Produit_Matrice_Scalaire;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul le produit d'une matrice--------------
   --------------et d'un vecteur et retourne donc un vecteur---------------
   -------------------------------------------------------------------------------
   FUNCTION Produit_Matrice_Vecteur (
         M1 : IN     Matrice;
         V1 : IN     VECTEUR)
     RETURN Vecteur IS
      V : Vecteur;
   BEGIN
      V.L := M1.L;
      IF(M1.C/=V1.L)THEN
         RAISE Erreur_Dimension_Produit_Matrice_Vecteur;
      ELSE
         FOR I IN 1 .. M1.L LOOP
            FOR K IN 1 .. M1.C LOOP
               V.VECT(I) := V.VECT(I) + M1.Mat(I,K)*V1.VECT(K);
            END LOOP;
         END LOOP;
      END IF;
      RETURN V;
   END Produit_Matrice_Vecteur;
   -------------------------------------------------------------------------------




   ---------------Cette fonction permet de rcuprer le---------------
   ----------------i me vecteur colonne d'une matrice----------------
   -------------------------------------------------------------------
   FUNCTION To_Vecteur_Colonne (
         M : IN     Matrice;
         I : IN     Integer)
     RETURN Vecteur IS
      V : Vecteur;
   BEGIN
      V.L := M.L;
      FOR J IN 1 .. M.L LOOP
         V.Vect(J) := M.Mat(J,I);
      END LOOP;
      RETURN V;
   END To_Vecteur_Colonne;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul le produit de deux--------------
   ----------------vecteurs et retourne donc un float----------------
   --------------------------------------------------------------------
   FUNCTION Produit_Vecteur (
         V1 : IN     VECTEUR;
         V2 : IN     VECTEUR)
     RETURN Float IS
      F : Float;
   BEGIN
      F := 0.0;
      IF(V1.L/=V2.L)THEN
         RAISE Erreur_Dimension_Produit_Vecteur;
      ELSE
         FOR I IN 1 .. V1.L LOOP
            F := F + V2.VECT(I)*V1.VECT(I);
         END LOOP;
      END IF;
      RETURN F;
   END Produit_Vecteur;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcul le somme de deux----------------
   ---------------vecteurs et retourne donc un vecteur-----------------
   --------------------------------------------------------------------
   FUNCTION Somme_Vecteur (
         V1 : Vecteur;
         V2 : Vecteur)
     RETURN Vecteur IS
      V : Vecteur;
   BEGIN

      IF(V1.L/=V2.L)THEN
         RAISE Erreur_Dimension_Somme_Vecteur;
      ELSE
         V.L := V1.L;
         FOR I IN 1 .. V1.L LOOP
            V.Vect(I) := V1.VEct(I) + V2.VECT(I);
         END LOOP;
      END IF;
      RETURN V;
   END Somme_Vecteur;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul le produit d'un scalaire--------------
   ------------------et d'un vecteur => on multiplie chaque -----------------
   ------------------------composante par le scalaire------------------------
   --------------------------------------------------------------------------
   FUNCTION Produit_Vecteur_Scalaire (
         V : Vecteur;
         S : Float)
     RETURN Vecteur IS
      Vr : Vecteur;
   BEGIN
      Vr.L := V.L;
      FOR I IN 1 .. V.L LOOP
         Vr.Vect(I) := V.Vect(I)*S;
      END LOOP;
      RETURN Vr;
   END Produit_Vecteur_Scalaire;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul la norme d'un--------------
   --------------vecteur et retourne donc un float----------------
   ---------------------------------------------------------------
   FUNCTION Norme_Vecteur (
         V : IN     Vecteur)
     RETURN FLOAT IS
      Res : Float := 0.0;
   BEGIN
      FOR I IN 1 .. V.L LOOP
         Res := Res + V.VECT(I)**2;
      END LOOP;
      RETURN Sqrt(Res);
   END Norme_Vecteur;
   -------------------------------------------------------------------------------




   --------------Cette fonction calcul la distance entre---------------
   --------------deux vecteurs et retourne donc un float---------------
   --------------------------------------------------------------------
   FUNCTION Distance (
         V1 : IN     VECTEUR;
         V2 : IN     VECTEUR)
     RETURN Float IS
      Res : Float := 0.0;
   BEGIN
      IF(V1.L /= V2.L)THEN
         RAISE Erreur_Dimension_Distance_Vecteur;
      ELSE
         FOR I IN 1 .. V1.L LOOP
            Res := Res + (V1.Vect(I)-V2.VECT(I))**2;
         END LOOP;
      END IF;
      RETURN Sqrt(Res);
   END Distance;
   -------------------------------------------------------------------------------


   --------------Cette fonction dfinie l'quation de rcurrence--------------
   -------------- partir de Vn on calcule Vn+1 avec la matrice---------------
   -------------------------------------------------------------------------------
   FUNCTION Calcul_Recurrence (
         Vi : IN     VECTEUR;
         M  : IN     Matrice)
     RETURN Vecteur IS
      Vs : VECTEUR;
   BEGIN
      Vs := Produit_Matrice_Vecteur(M,Vi);
      Vs := Produit_Vecteur_Scalaire(Vs,1.0/Norme_Vecteur(Vs));

      RETURN Vs;

   END Calcul_Recurrence;
   -------------------------------------------------------------------------------


   --------------Cette fonction calcule un vecteur propre unitaire--------------
   --------------associ  la plus grande valeur propre de la matrice-----------
   --------------On rpte l'opration de recurrence juqu' ce qu'on ait--------
   -----------------------convergence au sens de Cauchy-------------------------
   -----------------------------------------------------------------------------
   FUNCTION Calcul_Vect_Propre_Max (
         M : IN     Matrice)
     RETURN Vecteur IS
      Vn : Vecteur;
      Vm : Vecteur;
   BEGIN
      Vn.L := M.C;
      Vm.L := M.C;

      FOR I IN 1 .. Vn.L LOOP
         Vn.Vect(I) := 1.0;-- Initiallisation du vecteur
      END LOOP;
      Vn := Produit_Vecteur_Scalaire(Vn,1.0/Norme_Vecteur(Vn)); -- Normalisation du vecteur : il est unitaire

      Vm := Calcul_Recurrence(Vn,M);
      WHILE(Distance(Vn,Vm)>Valconv)LOOP
         Vn := Vm;
         Vm := Calcul_Recurrence(Vn,M);
      END LOOP;
      RETURN Vm;
   END Calcul_Vect_Propre_Max;
   -------------------------------------------------------------------------------


   --------------Cette fonction rcupre la position de la composante--------------
   -------------d'un vecteur qui possde la plus grande valeur absolue-------------
   --------------------------------------------------------------------------------
   FUNCTION Composante_Maximum (
         V : IN     Vecteur)
     RETURN Integer IS
      Pos : Integer := 1;
   BEGIN
      FOR I IN 2 .. V.L LOOP
         IF (ABS(V.Vect(I))>V.Vect(Pos)) THEN
            Pos := I;
         END IF;
      END LOOP;
      RETURN Pos;
   END Composante_Maximum;
   -------------------------------------------------------------------------------


   --------------Cette fonction permet de rcuprer la plus grande------------
   ----------------valeur propre de notre matrice. On utilise le--------------
   -------------------vecteur propre calcul prcdement.---------------------
   ---------------------------------------------------------------------------
   FUNCTION Calcul_Val_Propre_Max (
         V : IN     Vecteur;
         M : IN     Matrice)
     RETURN Float IS
      Vim : Vecteur;
      Pos : Integer := 0;
   BEGIN
      Vim := Produit_Matrice_Vecteur(M,V);

      Pos := Composante_Maximum(V);

      RETURN Vim.Vect(Pos)/V.Vect(Pos);
   END Calcul_Val_Propre_Max;
   -------------------------------------------------------------------------------




   ---------------Cette fonction permet de caluler la---------------
   ---------------composante principale associe  un---------------
   --------------vecteur propre. On projte donc chaque-------------
   --------------vecteur colonne sur le vecteur propre--------------
   ------------et on retourne la somme de ces projections-----------
   -----------------------------------------------------------------
   FUNCTION Calcul_Composante_Principale (
         Vect_Propre : IN     Vecteur;
         M           : IN     Matrice)
     RETURN Vecteur IS
      Vect_Comp_Principale : Vecteur;
      TMP                  : Vecteur;
   BEGIN
      Vect_Comp_Principale.L := Vect_Propre.L;
      FOR I IN 1 .. M.C LOOP
         TMP := To_Vecteur_Colonne(M,I); -- on rcupre la i me colonne
         TMP := Produit_Vecteur_Scalaire(TMP,Produit_Vecteur(TMP,Vect_Propre)); -- on calcule la projection de la ime
         -- colonne sur le vecteur propre : c'est
         -- k fois le vecteur propre (unitaire),
         -- avec k = produit scalaire du vecteur propre
         -- et du vecteur colonne.
         Vect_Comp_Principale := Somme_Vecteur(Vect_Comp_Principale,TMP); -- On somme ces projections
      END LOOP;

      RETURN Vect_Comp_Principale;
   END Calcul_Composante_Principale;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcule une matrice transforme------------
   --------------afin que la plus grande valeur propre soit ---------------
   --------------l'ancienne 2me plus grande valeur propre-----------------
   ------------------------------------------------------------------------
   FUNCTION Deflation (
         M      : IN     Matrice;
         V      : IN     Vecteur;
         Lambda : IN     Float)
      RETURN Matrice IS
      U : Matrice;
   BEGIN
      U.L := M.L;
      U.C := M.C;
      FOR I IN 1 .. U.L LOOP
         FOR J IN 1 .. U.C LOOP
            U.Mat(I,J) := V.Vect(I) * V.Vect(J);
         END LOOP;
      END LOOP;

      U := Produit_Matrice_Scalaire(U,-Lambda);
      U := Somme_Matrice(M,U);
      RETURN U;
   END Deflation;
   -------------------------------------------------------------------------------



   --------------Cette fonction calcule les valeurs propres------------
   --------------les vecteurs propres associs (unitaires) ------------
   --------------ainsi que les composantes principales associes.------
   --------------------------------------------------------------------
   FUNCTION Calcul_Resultat (
         M : IN     Matrice)
     RETURN T_Resultat IS
      Tab : T_Resultat (1 .. M.C);
      U   : Matrice;
   BEGIN
      U := M;
      FOR I IN 1 .. U.C LOOP
         Tab(I).Vect_Propre := Calcul_Vect_Propre_Max(U); -- Calcul d'un vecteur propre associ  la plus grande valeur propre

         Tab(I).Val_Propre := Calcul_Val_Propre_Max(Tab(I).Vect_Propre,U); -- Calcul de la valeur propre correspondante au vecteur propre

         Tab(I).Composante_Principale := Calcul_Composante_Principale(Tab(I).Vect_Propre,M); -- Calcul de la composante principale

         U := Deflation(U,Tab(I).Vect_Propre,Tab(I).Val_Propre); -- Dflation de la matrice afin que la nouvelle plus grande valeur propre
         -- soit l'ancienne deuxime plus grande valeur propre.
      END LOOP;
      RETURN Tab;
   END Calcul_Resultat;
   -------------------------------------------------------------------------------



   -----------------------------Affichage------------------------------------
   --------------------------------------------------------------------------
   PROCEDURE Affichage (
         Tab : IN     T_Resultat;
         N   : IN     Integer) IS
   BEGIN
      FOR I IN 1 .. N LOOP
         Put("Valeur propre " & Integer'Image(I) & " : " );
         Put(Tab(I).Val_Propre,0,0,0);
         New_Line;
         Put("Vecteur propre unitaire associ : [ " );
         Put(Tab(I).Vect_Propre.Vect(1),0,4,0);
         FOR J IN 2 .. N LOOP
            Put(" ; ");
            Put(Tab(I).Vect_Propre.Vect(J),0,4,0);
         END LOOP;
         Put_Line(" ]");
         New_Line;
         Put("Composante principale associ : [ " );
         Put(Tab(I).Composante_Principale.Vect(1),0,4,0);
         FOR J IN 2 .. N LOOP
            Put(" ; ");
            Put(Tab(I).Composante_Principale.Vect(J),0,4,0);
         END LOOP;
         Put_Line(" ]");
         New_Line;
         New_Line;
      END LOOP;
   END Affichage;
   -------------------------------------------------------------------------------


END Fctmulti;
