WITH Const;
USE Const;

PACKAGE Fctmulti IS

   -- Dfinition des erreurs que l'on peut rencontrer

   Erreur_Dimension_Produit_Matrice,
   Erreur_Dimension_Somme_Matrice,
   Erreur_Dimension_Produit_Matrice_Vecteur,
   Erreur_Dimension_Produit_Vecteur,
   Erreur_Dimension_Somme_Vecteur,
   Erreur_Dimension_Distance_Vecteur        : EXCEPTION;

   -- Dfinition des types uitiliss
   TYPE T_MATRICE IS ARRAY (1 .. LigneMax, 1 .. ColonneMax) OF Float;

   TYPE Matrice IS
      RECORD
         Mat : T_Matrice := (OTHERS => (OTHERS => 0.0));
         L   : Integer   := 0;
         C   : Integer   := 0;
      END RECORD;

   TYPE T_VECTEUR IS ARRAY (1 .. LigneMax) OF Float;

   TYPE Vecteur IS
      RECORD
         Vect : T_VECTEUR := (OTHERS => 0.0);
         L    : Integer   := 0;
      END RECORD;


   TYPE Resultat IS
      RECORD
         Composante_Principale : Vecteur;
         Vect_Propre           : Vecteur;
         Val_Propre            : Float;
      END RECORD;

   TYPE T_Resultat IS ARRAY (Positive RANGE <>) OF Resultat;



   -- Dbut de la dfinition des fonctions


   --***Fonctions de crations***--
   FUNCTION Creer_Matrice (
         L : IN     Integer;
         C : IN     Integer)
     RETURN MATRICE;

   FUNCTION Creer_Vecteur (
         L : IN     Integer)
     RETURN Vecteur;
   --***Calculs sur les matrices***--
   FUNCTION Moyenne_Colonne (
         M : IN     Matrice)
     RETURN Vecteur;

   FUNCTION Moyenne_Carre_Colonne (
         M : IN     Matrice)
     RETURN Vecteur;

   FUNCTION Transposee (
         M : IN     Matrice)
     RETURN Matrice;

   FUNCTION Matrice_Centree_Reduite (
         M : IN     Matrice)
     RETURN Matrice;


   FUNCTION Produit_Matrice (
         M1 : IN     Matrice;
         M2 : IN     MATRICE)
     RETURN Matrice;

   FUNCTION Somme_Matrice (
         M1 : IN     Matrice;
         M2 : IN     MATRICE)
     RETURN Matrice;


   FUNCTION Produit_Matrice_Scalaire (
         M : Matrice;
         S : Float)
     RETURN Matrice;


   FUNCTION Produit_Matrice_Vecteur (
         M1 : IN     Matrice;
         V1 : IN     VECTEUR)
     RETURN Vecteur;

   FUNCTION To_Vecteur_Colonne (
         M : IN     Matrice;
         I : IN     Integer)
     RETURN Vecteur;

   --***Calculs sur les vecteurs***--
   FUNCTION Produit_Vecteur (
         V1 : IN     VECTEUR;
         V2 : IN     VECTEUR)
     RETURN Float;

   FUNCTION Somme_Vecteur (
         V1 : Vecteur;
         V2 : Vecteur)
     RETURN Vecteur;

   FUNCTION Produit_Vecteur_Scalaire (
         V : Vecteur;
         S : Float)
     RETURN Vecteur;


   FUNCTION Norme_Vecteur (
         V : IN     Vecteur)
     RETURN FLOAT;


   FUNCTION Distance (
         V1 : IN     VECTEUR;
         V2 : IN     VECTEUR)
     RETURN Float;

   --***Fonctions utilises pour caluler le vecteur propre max***--
   FUNCTION Calcul_Recurrence (
         Vi : IN     VECTEUR;
         M  : IN     Matrice)
     RETURN Vecteur;

   FUNCTION Calcul_Vect_Propre_Max (
         M : IN     Matrice)
     RETURN Vecteur;
   --***Fonctions utilises pour caluler la valeur propre max***--
   FUNCTION Composante_Maximum (
         V : IN     Vecteur)
     RETURN Integer;

   FUNCTION Calcul_Val_Propre_Max (
         V : IN     Vecteur;
         M : IN     Matrice)
     RETURN Float;

   --***Fonction utilise pour caluler la composante principale***--
   FUNCTION Calcul_Composante_Principale (
         Vect_Propre : IN     Vecteur;
         M           : IN     Matrice)
     RETURN Vecteur;
   --***Fonction qui transforme la matrice : nouvelle plus grande val propre = ancienne deuxime plus grande***--
   FUNCTION Deflation (
         M      : IN     Matrice;
         V      : IN     Vecteur;
         Lambda : IN     Float)
     RETURN Matrice;
   --***Fonction qui centralise le calcul des rsultats***--
   FUNCTION Calcul_Resultat (
         M : IN     Matrice)
     RETURN T_Resultat;
   --***Fonction d'affichage***--
   PROCEDURE Affichage (
         Tab : IN     T_Resultat;
         N   : IN     Integer);




END Fctmulti;